/*
 * Copyright (c) 2016, Freescale Semiconductor, Inc.
 * Copyright 2016-2018 NXP
 * All rights reserved.
 *
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

#include "fsl_flexspi.h"
#include "app.h"
#include "fsl_debug_console.h"
#include "fsl_cache.h"

#include "pin_mux.h"
#include "board.h"
#include "clock_config.h"
#include "fsl_common.h"
/*******************************************************************************
 * Definitions
 ******************************************************************************/
/*******************************************************************************
 * Prototypes
 ******************************************************************************/

/*******************************************************************************
 * Variables
 ******************************************************************************/
static uint8_t s_nand_program_buffer[FLASH_PAGE_SIZE];
static uint8_t s_nand_read_buffer[FLASH_PAGE_SIZE];

extern status_t flexspi_nand_flash_erase_block(FLEXSPI_Type *base, uint32_t address);
extern status_t flexspi_nand_flash_page_program(FLEXSPI_Type *base, uint32_t dstAddr, const uint32_t *src);
extern status_t flexspi_nand_get_vendor_id(FLEXSPI_Type *base, uint8_t *vendorId);\
extern void flexspi_nand_flash_init(FLEXSPI_Type *base);
/*******************************************************************************
 * Code
 ******************************************************************************/
flexspi_device_config_t deviceconfig = {
    .flexspiRootClk       = 104000000, //133000000,
    .flashSize            = FLASH_SIZE,
    .CSIntervalUnit       = kFLEXSPI_CsIntervalUnit1SckCycle,
    .CSInterval           = 2,
    .CSHoldTime           = 3,
    .CSSetupTime          = 3,
    .dataValidTime        = 0,
    .columnspace          = 0,
    .enableWordAddress    = 0,
    .AWRSeqIndex          = 0,
    .AWRSeqNumber         = 0,
    .ARDSeqIndex          = NAND_CMD_LUT_SEQ_IDX_READ_FROM_CACHE,
    .ARDSeqNumber         = 1,
    .AHBWriteWaitUnit     = kFLEXSPI_AhbWriteWaitUnit2AhbCycle,
    .AHBWriteWaitInterval = 0,
};


const uint32_t customLUT[CUSTOM_LUT_LENGTH] = {
	// Read from Cache:
	[4 * NAND_CMD_LUT_SEQ_IDX_READ_FROM_CACHE] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x6B, kFLEXSPI_Command_DUMMY_SDR, kFLEXSPI_1PAD, 3),
	[4 * NAND_CMD_LUT_SEQ_IDX_READ_FROM_CACHE + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_CADDR_SDR, kFLEXSPI_1PAD, 13, kFLEXSPI_Command_DUMMY_SDR, kFLEXSPI_4PAD, 8),
	[4 * NAND_CMD_LUT_SEQ_IDX_READ_FROM_CACHE + 2] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_READ_SDR, kFLEXSPI_4PAD, 0x80, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),

    /* Read Status */
    [4 * NAND_CMD_LUT_SEQ_IDX_READSTATUS] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x0F, kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0xC0),
	[4 * NAND_CMD_LUT_SEQ_IDX_READSTATUS + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_READ_SDR, kFLEXSPI_1PAD, 0x01, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),

	// Write Enable
	[4 * NAND_CMD_LUT_SEQ_IDX_WRITEENABLE] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x06, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),

	// Erase Sector
	[4 * NAND_CMD_LUT_SEQ_IDX_BLOCKERASE] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0xD8, kFLEXSPI_Command_DUMMY_SDR, kFLEXSPI_1PAD, 7),
	[4 * NAND_CMD_LUT_SEQ_IDX_BLOCKERASE + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_RADDR_SDR, kFLEXSPI_1PAD, 17, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),

	// Page Program Load
	[4 * NAND_CMD_LUT_SEQ_IDX_PROGRAMLOAD] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x02, kFLEXSPI_Command_DUMMY_SDR, kFLEXSPI_1PAD, 3),
	[4 * NAND_CMD_LUT_SEQ_IDX_PROGRAMLOAD + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_CADDR_SDR, kFLEXSPI_1PAD, 13, kFLEXSPI_Command_WRITE_SDR, kFLEXSPI_1PAD, 0x10),
	[4 * NAND_CMD_LUT_SEQ_IDX_PROGRAMLOAD + 2] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0, 0, 0, 0),

	// Read Page (read page INTO cache   (Read Cell Array) ):
	[4 * NAND_CMD_LUT_SEQ_IDX_READ_PAGE] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x13, kFLEXSPI_Command_DUMMY_SDR, kFLEXSPI_1PAD, 7),
	[4 * NAND_CMD_LUT_SEQ_IDX_READ_PAGE + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_RADDR_SDR, kFLEXSPI_1PAD, 17, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),


	// Read Status Register (including ECC status)
	[4 * NAND_CMD_LUT_SEQ_IDX_READ_ECC_STATUS] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x0F, kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0xC0),
	[4 * NAND_CMD_LUT_SEQ_IDX_READ_ECC_STATUS + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_READ_SDR, kFLEXSPI_1PAD, 8, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),

	// Page Program Execute
	[4 * NAND_CMD_LUT_SEQ_IDX_PROGRAM_EXECUTE] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x10, kFLEXSPI_Command_DUMMY_SDR, kFLEXSPI_1PAD, 7),
	[4 * NAND_CMD_LUT_SEQ_IDX_PROGRAM_EXECUTE + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_RADDR_SDR, kFLEXSPI_1PAD, 17, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),

	//Clear for TC58CVG2S0H - NAND FLASH:
	/* Read ID */
	[4 * NAND_CMD_LUT_SEQ_IDX_READID] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_SDR, kFLEXSPI_1PAD, 0x9F, kFLEXSPI_Command_DUMMY_SDR, kFLEXSPI_1PAD, 8),
	[4 * NAND_CMD_LUT_SEQ_IDX_READID + 1] = FLEXSPI_LUT_SEQ(kFLEXSPI_Command_READ_SDR, kFLEXSPI_1PAD, 16, kFLEXSPI_Command_STOP, kFLEXSPI_1PAD, 0),

};


int main(void)
{
    uint32_t i = 0;
    uint32_t flashBaseAddress = EXAMPLE_FLEXSPI_AMBA_BASE + EXAMPLE_BLOCK_TO_ERASE * BLOCK_SIZE;
    status_t status;
    uint8_t vendorID[4];

    BOARD_ConfigMPU();
    BOARD_InitPins();
    BOARD_BootClockRUN();
//    BOARD_InitDebugConsole();
    SCB_DisableDCache();

    flexspi_nand_flash_init(EXAMPLE_FLEXSPI);

//    PRINTF("\r\nFLEXSPI example started!\r\n");

    /* Get vendor ID. */
    status = flexspi_nand_get_vendor_id(EXAMPLE_FLEXSPI, vendorID);
    if (status != kStatus_Success)
    {
        return status;
    }
//    PRINTF("Vendor ID: 0x%x, Device ID: 0x%x\r\n", vendorID[0], vendorID[0]);


    /* Erase sectors. */
//    PRINTF("Erasing Serial NAND over FlexSPI...\r\n");
    status = flexspi_nand_flash_erase_block(EXAMPLE_FLEXSPI, EXAMPLE_BLOCK_TO_ERASE * BLOCK_SIZE);
    if (status != kStatus_Success)
    {
        PRINTF("Erase sector failure !\r\n");
        return -1;
    }

    memset(s_nand_program_buffer, 0xFFU, sizeof(s_nand_program_buffer));

//    delay_ms(5);

    DCACHE_CleanInvalidateByRange(EXAMPLE_FLEXSPI_AMBA_BASE + EXAMPLE_PAGE_TO_PROGRAM_READ * FLASH_PAGE_SIZE, FLASH_PAGE_SIZE);
    memcpy(s_nand_read_buffer, (void *)(flashBaseAddress), sizeof(s_nand_read_buffer));

    if (memcmp(s_nand_program_buffer, s_nand_read_buffer, sizeof(s_nand_program_buffer))) {
         PRINTF("Erase data -  read out data value incorrect !\r\n ");
        return -1;
    } else {
//        PRINTF("Erase data - successfully. \r\n");
    }

    for (i = 0; i < 0xFFU; i++) {
    	s_nand_program_buffer[i] = i;
    }

    status = flexspi_nand_flash_page_program(EXAMPLE_FLEXSPI, EXAMPLE_PAGE_TO_PROGRAM_READ * FLASH_PAGE_SIZE, (void *)s_nand_program_buffer);
    if (status != kStatus_Success) {
        PRINTF("Page program failure !\r\n");
        return -1;
    }

    DCACHE_CleanInvalidateByRange(EXAMPLE_FLEXSPI_AMBA_BASE + EXAMPLE_PAGE_TO_PROGRAM_READ * FLASH_PAGE_SIZE, FLASH_PAGE_SIZE);

    memcpy(s_nand_read_buffer, (void *)(EXAMPLE_FLEXSPI_AMBA_BASE + EXAMPLE_PAGE_TO_PROGRAM_READ * FLASH_PAGE_SIZE), sizeof(s_nand_read_buffer));

    if (memcmp(s_nand_read_buffer, s_nand_program_buffer, sizeof(s_nand_program_buffer)) != 0) {
        PRINTF("Program data -  read out data value incorrect !\r\n ");
        return -1;
    } else {
        PRINTF("Program data - successfully. \r\n");
    }

    while (1)
    {
    }
}
